
#ifdef 		MODVERSIONS
    #ifndef 	MODULE
        #define 	MODULE
    #endif
#endif

#include <linux/version.h>
#define	VERSION_CODE(ver,rel,seq)	((ver << 16) | (rel << 8) | seq)

#ifdef MODULE
    #ifdef MODVERSIONS
        #include <linux/modversions.h>
    #endif
    #include <linux/module.h>
#else
    #define	MOD_INC_USE_COUNT
    #define MOD_DEC_USE_COUNT
#endif



#define DRIVER_NAME         "PTLSER : "


#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/delay.h>

#if (LINUX_VERSION_CODE < VERSION_CODE(3,4,0))
#include <asm/system.h>
#endif
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/bitops.h>

#include "ptlser.h"
#include "register.h"
#ifdef CONFIG_PCI
    #include <linux/pci.h>
#endif

#define	PTLSER_VERSION	"1.04"
#define	PTLSERMAJOR	31
#define	PTLSERCUMAJOR	34



enum	{
	PTLSER_PCIE_PORT_8 = 1,
	PTLSER_PCIE_PORT_4,
	PTLSER_PCIE_PORT_2,
	PTLSER_PCIE_PORT_1,
	PTLSER_PCI_PORT_8,
	PTLSER_PCI_PORT_4,
	PTLSER_PCI_PORT_2,
	PTLSER_PCI_PORT_1,

};

static char *ptlser_brdname[] = {
	"Pericom PI7C9X7958 PCI Express UART Device (8 ports)",
	"Pericom PI7C9X7954 PCI Express UART Device (4 ports)",
	"Pericom PI7C9X7952 PCI Express UART Device (2 ports)",
	"Pericom PI7C9X7951 PCI Express UART Device (1 port)",
	"Pericom PI7C9X8958 PCI UART Device (8 ports)",
	"Pericom PI7C9X8954 PCI UART Device (4 ports)",
	"Pericom PI7C9X8952 PCI UART Device (2 ports)",
	"Pericom PI7C9X8951 PCI UART Device (1 port)",
	
};

static int ptlser_numports[] = {
	8,
	4,
	2,
	1,	
	8,
	4,
	2,
	1,		
};


static  struct pci_device_id	ptlser_pcibrds[] = {
    {PTLSER_VENDOR_ID, PTLSER_DEVICE_7958, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PTLSER_PCIE_PORT_8},
    {PTLSER_VENDOR_ID, PTLSER_DEVICE_7954, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PTLSER_PCIE_PORT_4},
    {PTLSER_VENDOR_ID, PTLSER_DEVICE_7952, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PTLSER_PCIE_PORT_2},
    {PTLSER_VENDOR_ID, PTLSER_DEVICE_7951, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PTLSER_PCIE_PORT_1},
    {PTLSER_VENDOR_ID, PTLSER_DEVICE_8958, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PTLSER_PCI_PORT_8},
    {PTLSER_VENDOR_ID, PTLSER_DEVICE_8954, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PTLSER_PCI_PORT_4},
    {PTLSER_VENDOR_ID, PTLSER_DEVICE_8952, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PTLSER_PCI_PORT_2},
    {PTLSER_VENDOR_ID, PTLSER_DEVICE_8951, PCI_ANY_ID, PCI_ANY_ID, 0, 0, PTLSER_PCI_PORT_1},    
	{0}
};

MODULE_DEVICE_TABLE(pci, ptlser_pcibrds);

typedef struct _ptlser_pci_info {
	unsigned short busNum;
	unsigned short devNum;
struct pci_dev	*pdev;	
} ptlser_pci_info;

static int              ttymajor=PTLSERMAJOR;
static int              calloutmajor=PTLSERCUMAJOR;
static unsigned char    interface=0;
static unsigned char    terminator=0;


#ifdef MODULE
/* Variables for insmod */
MODULE_LICENSE("GPL");
MODULE_AUTHOR("simoncc");
MODULE_DESCRIPTION("Pericom PI7C9X795x/PI7C9X895x UART Device Driver");

int ptlser_ioaddr_array_num;
module_param(ttymajor, int, 0);
module_param(calloutmajor, int, 0);
module_param(interface, byte, 0);
module_param(terminator, byte, 0);

#endif /* MODULE */


struct ptlser_hwconf {
	int		        board_type;
	int		        ports;
	int		        irq;
	unsigned long	vector_mask;
	int		        uart_type;
	unsigned char	*ioaddr[PTLSER_PORTS_PER_BOARD];
	int		        baud_base[PTLSER_PORTS_PER_BOARD];
	ptlser_pci_info	pciInfo;
	int		        MaxCanSetBaudRate[PTLSER_PORTS_PER_BOARD];
	unsigned long	*opmode_ioaddr[PTLSER_PORTS_PER_BOARD];
};


struct ptlser_struct {
	
#if (LINUX_VERSION_CODE >= VERSION_CODE(3,7,0))
	struct tty_port ttyPort;
#endif
	int			        port;
	unsigned char		*base;		/* port base address */
	int			        irq;		/* port using irq no. */
	int			        baud_base;	/* max. speed */
	int			        flags;		/* defined in tty.h */
	int			        type;		/* UART type */
	struct tty_struct   *tty;
	int			        read_status_mask;
	int			        ignore_status_mask;
	int			        xmit_fifo_size;
	int			        custom_divisor;
	int			        close_delay;
	unsigned short		closing_wait;
	int			        IER;		/* Interrupt Enable Register */
	int			        MCR;		/* Modem control register */
	unsigned long		event;
	int			        count;		/* # of fd on device */
	int			        blocked_open;	/* # of blocked opens */
	unsigned char		*xmit_buf;
	int			        xmit_head;
	int			        xmit_tail;
	int			        xmit_cnt;
	struct work_struct tqueue;
	struct ktermios		normal_termios;
	struct ktermios		callout_termios;
	wait_queue_head_t   open_wait;
	wait_queue_head_t   close_wait;
	wait_queue_head_t   delta_msr_wait;
	struct async_icount	icount; 	/* kernel counters for the 4 input interrupts */
	int			        timeout;
	int			        MaxCanSetBaudRate;	
	long                realbaud;
//	unsigned char       err_shadow;
	spinlock_t		    slock;
	int			        speed;
	int			        custom_baud_rate;
	int			        board_type;
	unsigned char 		intr_status;
	unsigned char		lsr_status;
};


static struct tty_driver	*ptlser_var_sdriver;

static struct ktermios      *ptlser_var_termios[PTLSER_PORTS+1];
#if (LINUX_VERSION_CODE < VERSION_CODE(3,3,0))
static struct ktermios 		*ptlser_var_termios_locked[PTLSER_PORTS+1];
#endif

static struct ptlser_struct	ptlser_var_table[PTLSER_PORTS];
static struct tty_struct    *ptlser_var_tty[PTLSER_PORTS+1];


/*
 * This is used to figure out the divisor speeds and the timeouts
 */

static int ptlser_baud_table[] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600, 0 };
#define BAUD_TABLE_NO	(sizeof(ptlser_baud_table)/sizeof(int))

struct ptlser_hwconf ptlsercfg[PTLSER_BOARDS];

/*
 * static functions:
 */

#ifdef MODULE
int		init_module(void);
void	cleanup_module(void);
#endif

static int	ptlser_open(struct tty_struct *, struct file *);
static void	ptlser_close(struct tty_struct *, struct file *);
static int	ptlser_write(struct tty_struct *, const unsigned char *, int);
static int	ptlser_put_char(struct tty_struct *, unsigned char);
static void	ptlser_flush_chars(struct tty_struct *);
static int	ptlser_write_room(struct tty_struct *);
static int	ptlser_chars_in_buffer(struct tty_struct *);
static void	ptlser_flush_buffer(struct tty_struct *);
static int 	ptlser_ioctl(struct tty_struct * tty, unsigned int cmd, unsigned long arg);
static void	ptlser_throttle(struct tty_struct *);
static void	ptlser_unthrottle(struct tty_struct *);
static void	ptlser_set_termios(struct tty_struct *, struct ktermios *);
static void	ptlser_stop(struct tty_struct *);
static void	ptlser_start(struct tty_struct *);
static void	ptlser_hangup(struct tty_struct *);
static int	ptlser_tiocmget(struct tty_struct *, struct file *);
static int	ptlser_tiocmset(struct tty_struct *, struct file *, unsigned int, unsigned int);
static int	ptlser_break_ctl(struct tty_struct *, int);
static void 	ptlser_wait_until_sent(struct tty_struct *tty, int timeout);

#ifdef CONFIG_PCI
static int      ptlser_get_PCI_conf(int ,int ,int ,struct ptlser_hwconf *);
#endif

static void     ptlser_do_softint(struct work_struct *work);
static irqreturn_t ptlser_interrupt(int irq, void *dev_id);
static void	ptlser_receive_chars(struct ptlser_struct *, int *);
static void	ptlser_transmit_chars(struct ptlser_struct *);
static void	ptlser_check_modem_status(struct ptlser_struct *, int);
static int	ptlser_block_til_ready(struct tty_struct *, struct file *, struct ptlser_struct *);
static int	ptlser_startup(struct ptlser_struct *);
static void	ptlser_shutdown(struct ptlser_struct *);
static int	ptlser_change_speed(struct ptlser_struct *, struct ktermios *old_termios);
static int	ptlser_get_serial_info(struct ptlser_struct *, struct serial_struct *);
static int	ptlser_set_serial_info(struct ptlser_struct *, struct serial_struct *);
static int	ptlser_get_lsr_info(struct ptlser_struct *, unsigned int *);
static void	ptlser_send_break(struct ptlser_struct *, int);
static int	ptlser_set_baud(struct ptlser_struct *info, long newspd);



static struct tty_operations ptlser_ops = {
	.open               = ptlser_open,
	.close              = ptlser_close,
	.write              = ptlser_write,
	.put_char           = ptlser_put_char,
	.flush_chars        = ptlser_flush_chars,
	.write_room         = ptlser_write_room,
	.chars_in_buffer    = ptlser_chars_in_buffer,
	.flush_buffer       = ptlser_flush_buffer,
	.ioctl              = ptlser_ioctl,
	.throttle           = ptlser_throttle,
	.unthrottle         = ptlser_unthrottle,
	.set_termios        = ptlser_set_termios,
	.stop               = ptlser_stop,
	.start              = ptlser_start,
	.hangup             = ptlser_hangup,
	.tiocmget           = ptlser_tiocmget,
	.tiocmset           = ptlser_tiocmset,
	.break_ctl          = ptlser_break_ctl,
	.wait_until_sent    = ptlser_wait_until_sent,

};


#ifdef CONFIG_PCI
static int ptlser_get_PCI_conf(int busnum, int devnum, int board_type, struct ptlser_hwconf *hwconf)
{
	int		i;
	unsigned char	*ioaddress;
	struct pci_dev	*pdev=hwconf->pciInfo.pdev;

	//io address
	hwconf->board_type = board_type;
	hwconf->ports = ptlser_numports[board_type-1];

	request_mem_region(pci_resource_start(pdev, 1), pci_resource_len(pdev, 1), "ptlser(MEM)");
	ioaddress = ioremap(pci_resource_start(pdev,1), pci_resource_len(pdev,1));
	
	for (i = 0; i < hwconf->ports; i++) {
		hwconf->ioaddr[i] = ioaddress + (i * PTLSER_MEMORY_RANGE);
		if(i == 3){
			switch(board_type){
			case PTLSER_PCIE_PORT_4:
				hwconf->ioaddr[i] = ioaddress + (PTLSER_PORT8 * PTLSER_MEMORY_RANGE);
				break;
			default:
				break;
			}
		}
	}

	//irq
	hwconf->irq = hwconf->pciInfo.pdev->irq;
	hwconf->uart_type = PORT_16550A;
	
    for (i = 0; i < hwconf->ports; i++) {
		hwconf->baud_base[i] = hwconf->MaxCanSetBaudRate[i] = 921600;
	}

	return 0;
}
#endif


static void ptlser_do_softint(struct work_struct *work)
{
        struct ptlser_struct *info = container_of(work, struct ptlser_struct, tqueue);
	struct tty_struct *	tty = info->tty;

	if ( test_and_clear_bit(PTLSER_EVENT_TXLOW, &info->event) ) 
		tty_wakeup(tty);
	if ( test_and_clear_bit(PTLSER_EVENT_HANGUP, &info->event) )
		tty_hangup(tty);
}

unsigned char getTxCount(struct ptlser_struct *	info)
{
	unsigned char value;
	value = PTLSER_READ_REG(info->base + PTLSER_LTF_OFFSET);
	if (value > PTLSER_FIFO_SIZE)
		value = PTLSER_FIFO_SIZE;
	return value;
}

unsigned char getRxCount(struct ptlser_struct *	info)
{
	unsigned char value;
	value = PTLSER_READ_REG(info->base + PTLSER_LRF_OFFSET);
	if (value > PTLSER_FIFO_SIZE)
		value = PTLSER_FIFO_SIZE;
	return value;
}


/*
 * This routine is called whenever a serial port is opened.  It
 * enables interrupts for a serial port, linking in its async structure into
 * the IRQ chain.   It also performs the serial-specific
 * initialization for the tty structure.
 */
static int ptlser_open(struct tty_struct * tty, struct file * filp)
{
	struct ptlser_struct *	info;
	int			retval, line;

	PTLSER_LOCK_INIT();
	
	line = PORTNO(tty);

	if ( line == PTLSER_PORTS )
		return 0;
	    
    	PTLSER_MOD_INC;
    	    
	if ( (line < 0) || (line > PTLSER_PORTS) )
		return -ENODEV;

	info = ptlser_var_table + line;

	if ( !info->base )
	    return -ENODEV;

	tty->driver_data = info;
	info->tty = tty;

	/*
	 * Start up serial port
	 */
	info->count++; 
	retval = ptlser_startup(info);

	if ( retval )
	    return retval;

	retval = ptlser_block_til_ready(tty, filp, info);

	if ( retval )
	    return retval;

	if (info->count == 1) {
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	    if ( PTLSER_TTY_DRV(subtype) == SERIAL_TYPE_NORMAL )
		*tty->termios = info->normal_termios;
	    else
		*tty->termios = info->callout_termios;
#else
	    if ( PTLSER_TTY_DRV(subtype) == SERIAL_TYPE_NORMAL )
		(*tty).termios = info->normal_termios;
	    else
		(*tty).termios = info->callout_termios;	
#endif
	    PTLSER_LOCK(&info->slock);
	    ptlser_change_speed(info, 0);
	    PTLSER_UNLOCK(&info->slock);
	}

#ifdef TTY_DONT_FLIP
	clear_bit(TTY_DONT_FLIP, &tty->flags); // since VERSION_CODE >= 2.6.18
#endif

/* unmark here for very high baud rate (ex. 921600 bps) used
*/
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	tty->low_latency = 1;
#endif
#endif
	return 0;
}


/*
 * This routine is called when the serial port gets closed.  First, we
 * wait for the last remaining data to be sent.  Then, we unlink its
 * async structure from the interrupt chain if necessary, and we free
 * that IRQ if nothing is left in the chain.
 */
static void ptlser_close(struct tty_struct * tty, struct file * filp)
{
	struct ptlser_struct * info = (struct ptlser_struct *)tty->driver_data;
	unsigned long timeout;
	unsigned char reg_flag;	

	PTLSER_LOCK_INIT();

	if ( PORTNO(tty) == PTLSER_PORTS )
	    return;

	if ( !info ){
	    PTLSER_MOD_DEC;

	    return;
	}

	PTLSER_LOCK(&info->slock);

	if ( tty_hung_up_p(filp) ) {
	    PTLSER_UNLOCK(&info->slock);
	    PTLSER_MOD_DEC;

	    return;
	}

#ifndef SP1
	if ( (tty->count == 1) && (info->count != 1) ) {
#else
	if ((atomic_read(&tty->count) == 1) && (info->count != 1)) {
#endif /* SP1 */
	    /*
	     * Uh, oh.	tty->count is 1, which means that the tty
	     * structure will be freed.  Info->count should always
	     * be one in these conditions.  If it's greater than
	     * one, we've got real problems, since it means the
	     * serial port won't be shutdown.
	     */
	    info->count = 1;
	}

	if ( --info->count < 0 ) {
	    info->count = 0;
	}

	if ( info->count ) {
	    PTLSER_UNLOCK(&info->slock);
	    PTLSER_MOD_DEC;

	    return;
	}

	info->flags |= ASYNC_CLOSING;
	PTLSER_UNLOCK(&info->slock);
	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	if ( info->flags & ASYNC_NORMAL_ACTIVE )
	    info->normal_termios = *tty->termios;

	if ( info->flags & ASYNC_CALLOUT_ACTIVE )
	    info->callout_termios = *tty->termios;
#else
	if ( info->flags & ASYNC_NORMAL_ACTIVE )
	    info->normal_termios = (*tty).termios;

	if ( info->flags & ASYNC_CALLOUT_ACTIVE )
	    info->callout_termios = (*tty).termios;
#endif
	/*
	 * Now we wait for the transmit buffer to clear; and we notify
	 * the line discipline to only process XON/XOFF characters.
	 */
	
	reg_flag = PTLSER_READ_REG(info->base + PTLSER_EFR_OFFSET);
	reg_flag &= ~PTLSER_EFR_AUTO_RTS;
	PTLSER_WRITE_REG(reg_flag, info->base + PTLSER_EFR_OFFSET);
	tty->closing = 1;

	if ( info->closing_wait != ASYNC_CLOSING_WAIT_NONE )
	    tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * At this point we stop accepting input.  To do this, we
	 * disable the receive line status interrupts, and tell the
	 * interrupt driver to stop checking the data ready bit in the
	 * line status register.
	 */
	info->IER &= ~UART_IER_RLSI;

	if ( info->flags & ASYNC_INITIALIZED ) {
		PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	    /*
	     * Before we drop DTR, make sure the UART transmitter
	     * has completely drained; this is especially
	     * important if there is a transmit FIFO!
	     */
		timeout = jiffies + HZ;
		while ( ptlser_chars_in_buffer(tty) ) {
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(5);
			set_current_state(TASK_RUNNING);

			if ( time_after(jiffies, timeout) )
				break;
		    }
	}

	ptlser_shutdown(info);
	ptlser_flush_buffer(tty);
	tty_ldisc_flush(tty);
	    
	tty->closing = 0;
	info->event = 0;
	info->tty = 0;

	if ( info->blocked_open ) {
		if ( info->close_delay ) {
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(info->close_delay);
			set_current_state(TASK_RUNNING);
		}

		wake_up_interruptible(&info->open_wait);
	}

	info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE | ASYNC_CLOSING);
	wake_up_interruptible(&info->close_wait);

	PTLSER_MOD_DEC;
}

static int ptlser_write(struct tty_struct * tty, const unsigned char * buf, int count)
{
	int c, total = 0;
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
 
	PTLSER_LOCK_INIT();

	if ( !tty || !info->xmit_buf  )
	    return(0);
	
	while ( 1 ) {
		c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1, SERIAL_XMIT_SIZE - info->xmit_head));
		if ( c <= 0 )
			break;
		memcpy(info->xmit_buf + info->xmit_head, buf, c);
		PTLSER_LOCK(&info->slock);    
		info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE - 1);
		info->xmit_cnt += c;
		PTLSER_UNLOCK(&info->slock);

		buf += c;
		count -= c;
		total += c;
	}

	if ( info->xmit_cnt && !tty->stopped ) {
        	    PTLSER_LOCK(&info->slock); 	
		    info->IER &= ~UART_IER_THRI	;
	            PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	            info->IER |= UART_IER_THRI;
	            PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	            PTLSER_UNLOCK(&info->slock);
	}
	
	return total;
}

static int ptlser_put_char(struct tty_struct * tty, unsigned char ch)
{
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	PTLSER_LOCK_INIT();

	if ( !tty || !info->xmit_buf )
		return 0;
	if ( info->xmit_cnt >= SERIAL_XMIT_SIZE - 1 )
		return 0;

	PTLSER_LOCK(&info->slock);
	info->xmit_buf[info->xmit_head++] = ch;
	info->xmit_head &= SERIAL_XMIT_SIZE - 1;
	info->xmit_cnt++;
	PTLSER_UNLOCK(&info->slock);

	if ( !tty->stopped ) {
        	if (!tty->hw_stopped) {
        		PTLSER_LOCK(&info->slock);
		        info->IER &= ~UART_IER_THRI	;
	                PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	    		info->IER |= UART_IER_THRI;
	    		PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	    		PTLSER_UNLOCK(&info->slock);
		}
	}
	return 1;
}


static void ptlser_flush_chars(struct tty_struct * tty)
{
	char fcr;
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	PTLSER_LOCK_INIT();

	if ( info->xmit_cnt <= 0 || tty->stopped || !info->xmit_buf)
		return;
	PTLSER_LOCK(&info->slock);
        info->IER &= ~UART_IER_THRI;
        PTLSER_WRITE_REG(info->IER, info->base + UART_IER);

	info->IER |= UART_IER_THRI;
	PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	PTLSER_UNLOCK(&info->slock);
}


static int ptlser_write_room(struct tty_struct * tty)
{
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	int	ret;

	ret = SERIAL_XMIT_SIZE - info->xmit_cnt - 1;
	if ( ret < 0 )
	    ret = 0;
	    
	return ret;
}


static int ptlser_chars_in_buffer(struct tty_struct * tty)
{
	int len;
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	unsigned char t_cnt;
	
	len = info->xmit_cnt;
	t_cnt = getTxCount(info);

	if(t_cnt)
		len+=(int)t_cnt;

	return len;
}


static void ptlser_flush_buffer(struct tty_struct * tty)
{
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	char fcr;
	PTLSER_LOCK_INIT();

	PTLSER_LOCK(&info->slock);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	PTLSER_UNLOCK(&info->slock);

	PTLSER_WRITE_REG(0xFF, info->base + 0x24);
	PTLSER_WRITE_REG(0x01, info->base + 0x25);
	while( getTxCount(info) ) {
		PTLSER_WRITE_REG(PTLSER_FCR_TX_FLUSH, info->base + UART_FCR);
		PTLSER_WRITE_REG(PTLSER_FCR_TX_FLUSH, info->base + UART_FCR);
	} 
	while( getRxCount(info) ) {
		PTLSER_WRITE_REG(PTLSER_FCR_RX_FLUSH, info->base + UART_FCR);
		PTLSER_WRITE_REG(PTLSER_FCR_RX_FLUSH, info->base + UART_FCR);
	} 
	PTLSER_WRITE_REG(0x00, info->base + 0x24);
	PTLSER_WRITE_REG(0x02, info->base + 0x25);

	wake_up_interruptible(&tty->write_wait);

	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc->ops->write_wakeup)
		tty_wakeup(tty);
}

static int ptlser_ioctl(struct tty_struct * tty, unsigned int cmd, unsigned long arg)
{
	int			error;
	struct ptlser_struct *	info = (struct ptlser_struct *)tty->driver_data;
	int			retval;
	struct async_icount	cprev, cnow;	    /* kernel counter temps */
	struct serial_icounter_struct *p_cuser;     /* user space */
	unsigned long 		templ;
	PTLSER_LOCK_INIT();

	if ( (cmd != TIOCGSERIAL) && (cmd != TIOCMIWAIT) &&
	     (cmd != TIOCGICOUNT) ) {

		if ( tty->flags & (1 << TTY_IO_ERROR) )
			return -EIO;
	}

	switch ( cmd ) {
		case TCSBRK:	/* SVID version: non-zero arg --> no break */
			retval = tty_check_change(tty);

			if ( retval )
				return retval;

			tty_wait_until_sent(tty, 0);

			if ( !arg )
				ptlser_send_break(info, HZ/4);		/* 1/4 second */

			return 0;
		case TCSBRKP:	/* support for POSIX tcsendbreak() */
			retval = tty_check_change(tty);

			if ( retval )
				return retval;

			tty_wait_until_sent(tty, 0);
			ptlser_send_break(info, arg ? arg*(HZ/10) : HZ/4);

			return 0;
		case TIOCGSOFTCAR:
			error = PTLSER_ACCESS_CHK(VERIFY_WRITE, (void *)arg, sizeof(long));

			if ( PTLSER_ERR(error) )
				return error;

			put_to_user(C_CLOCAL(tty) ? 1 : 0, (unsigned long *)arg);

			return 0;
		case TIOCSSOFTCAR:
			error = PTLSER_ACCESS_CHK(VERIFY_READ, (void *)arg, sizeof(long));

			if ( PTLSER_ERR(error) )
				return error;

			get_from_user(templ,(unsigned long *)arg);
			arg = templ;
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))				
			tty->termios->c_cflag = ((tty->termios->c_cflag & ~CLOCAL) | (arg ? CLOCAL : 0));
#else
			tty->termios.c_cflag = ((tty->termios.c_cflag & ~CLOCAL) | (arg ? CLOCAL : 0));
#endif
			return 0;
		case TIOCGSERIAL:
			error = PTLSER_ACCESS_CHK(VERIFY_WRITE, (void *)arg,sizeof(struct serial_struct));

			if ( PTLSER_ERR(error) )
				return error;

			return ptlser_get_serial_info(info, (struct serial_struct *)arg);
		case TIOCSSERIAL:
			error = PTLSER_ACCESS_CHK(VERIFY_READ, (void *)arg,sizeof(struct serial_struct));

			if ( PTLSER_ERR(error) )
				return error;

			return ptlser_set_serial_info(info, (struct serial_struct *)arg);
		case TIOCSERGETLSR: /* Get line status register */
			error = PTLSER_ACCESS_CHK(VERIFY_WRITE, (void *)arg,sizeof(unsigned int));

			if ( PTLSER_ERR(error) )
				return error;
			else
				return ptlser_get_lsr_info(info, (unsigned int *)arg);
	/*
	 * Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change
	 * - mask passed in arg for lines of interest
	 *   (use |'ed TIOCM_RNG/DSR/CD/CTS for masking)
	 * Caller should use TIOCGICOUNT to see which one it was
	 */
		case TIOCMIWAIT:{
			DECLARE_WAITQUEUE(wait, current);
			int ret;

			PTLSER_LOCK(&info->slock);
			cprev = info->icount;   /* note the counters on entry */
			PTLSER_UNLOCK(&info->slock);
	    
			add_wait_queue(&info->delta_msr_wait, &wait);
			while ( 1 ) {
				PTLSER_LOCK(&info->slock);
				cnow = info->icount;	/* atomic copy */
				PTLSER_UNLOCK(&info->slock);
				set_current_state(TASK_INTERRUPTIBLE);

				if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
				     ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
                                     ((arg & TIOCM_CD) && (cnow.dcd != cprev.dcd)) ||
				     ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
					ret = 0;
					break;
				}
		/* see if a signal did it */
				if ( signal_pending(current) ){
					ret = -ERESTARTSYS;
					break;
				}

				cprev = cnow;
			}

			set_current_state(TASK_RUNNING);
			remove_wait_queue(&info->delta_msr_wait, &wait);
			break;
		}
	    /* NOTREACHED */
	/*
	 * Get counter of input serial line interrupts (DCD,RI,DSR,CTS)
	 * Return: write counters to the user passed counter struct
	 * NB: both 1->0 and 0->1 transitions are counted except for
	 *     RI where only 0->1 is counted.
	 */
		case TIOCGICOUNT:
			error = PTLSER_ACCESS_CHK(VERIFY_WRITE, (void *)arg,sizeof(struct serial_icounter_struct));

			if ( PTLSER_ERR(error) )
				return error;

			PTLSER_LOCK(&info->slock);
			cnow = info->icount;
			PTLSER_UNLOCK(&info->slock);
			p_cuser = (struct serial_icounter_struct *)arg;

			if (put_user(cnow.frame, &p_cuser->frame))
				return -EFAULT;
			
			if (put_user(cnow.brk, &p_cuser->brk))
				return -EFAULT;

			if (put_user(cnow.overrun, &p_cuser->overrun))
				return -EFAULT;

			if (put_user(cnow.buf_overrun, &p_cuser->buf_overrun))
				return -EFAULT;

			if (put_user(cnow.parity, &p_cuser->parity))
				return -EFAULT;

			if (put_user(cnow.rx, &p_cuser->rx))
				return -EFAULT;

			if (put_user(cnow.tx, &p_cuser->tx))
				return -EFAULT;

			put_to_user(cnow.cts, &p_cuser->cts);
			put_to_user(cnow.dsr, &p_cuser->dsr);
			put_to_user(cnow.rng, &p_cuser->rng);
			put_to_user(cnow.dcd, &p_cuser->dcd);

			return 0;

		default:
			return(-ENOIOCTLCMD);
	}

	return 0;
}



/*
 * This routine is called by the upper-layer tty layer to signal that
 * incoming characters should be throttled.
 */
static void ptlser_throttle(struct tty_struct * tty)
{
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	PTLSER_LOCK_INIT();
	PTLSER_LOCK(&info->slock);
	info->IER &= ~UART_IER_RDI;
	PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	PTLSER_UNLOCK(&info->slock);        
}


static void ptlser_unthrottle(struct tty_struct * tty)
{
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	PTLSER_LOCK_INIT();
	PTLSER_LOCK(&info->slock);
	info->IER |= UART_IER_RDI;
	PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	PTLSER_UNLOCK(&info->slock);        
}


static void ptlser_set_termios(struct tty_struct * tty, struct ktermios * old_termios)
{
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	PTLSER_LOCK_INIT();
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	if ( (tty->termios->c_cflag != old_termios->c_cflag) ||
	     (RELEVANT_IFLAG(tty->termios->c_iflag) != RELEVANT_IFLAG(old_termios->c_iflag)) ) {
		PTLSER_LOCK(&info->slock);
		ptlser_change_speed(info, old_termios);
		PTLSER_UNLOCK(&info->slock);

		if ( (old_termios->c_cflag & CRTSCTS) && !(tty->termios->c_cflag & CRTSCTS) ) {
	    		tty->hw_stopped = 0;
	    		ptlser_start(tty);
		}
	}
#else
	if ( (tty->termios.c_cflag != old_termios->c_cflag) ||
	     (RELEVANT_IFLAG(tty->termios.c_iflag) != RELEVANT_IFLAG(old_termios->c_iflag)) ) {
		PTLSER_LOCK(&info->slock);
		ptlser_change_speed(info, old_termios);
		PTLSER_UNLOCK(&info->slock);

		if ( (old_termios->c_cflag & CRTSCTS) && !(tty->termios.c_cflag & CRTSCTS) ) {
	    		tty->hw_stopped = 0;
	    		ptlser_start(tty);
		}
	}
#endif


/* Handle sw stopped */
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	if ( (old_termios->c_iflag & IXON) && !(tty->termios->c_iflag & IXON) ) {
#else
	if ( (old_termios->c_iflag & IXON) && !(tty->termios.c_iflag & IXON) ) {
#endif
    		tty->stopped = 0;
    		ptlser_start(tty);
	}
}


/*
 * ptlser_stop() and ptlser_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 */
static void ptlser_stop(struct tty_struct * tty)
{
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	PTLSER_LOCK_INIT();
	PTLSER_LOCK(&info->slock);
	if ( info->IER & UART_IER_THRI ) {
		info->IER &= ~UART_IER_THRI;
		PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	}
	PTLSER_UNLOCK(&info->slock);
}


static void ptlser_start(struct tty_struct * tty)
{
	struct ptlser_struct *info = (struct ptlser_struct *)tty->driver_data;
	PTLSER_LOCK_INIT();
	PTLSER_LOCK(&info->slock);
	if ( info->xmit_cnt && info->xmit_buf ) {
		info->IER &= ~UART_IER_THRI;
		PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
		info->IER |= UART_IER_THRI;
		PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
	}
	PTLSER_UNLOCK(&info->slock);
}

/*
 * ptlser_wait_until_sent() --- wait until the transmitter is empty
 */
#if (LINUX_VERSION_CODE >= 131394)
static void ptlser_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct ptlser_struct * info = (struct ptlser_struct *)tty->driver_data;
	unsigned long orig_jiffies, char_time;
	int lsr;
	if (info->type == PORT_UNKNOWN)
		return;

	if (info->xmit_fifo_size == 0)
		return; /* Just in case.... */

	orig_jiffies = jiffies;
	/*
	 * Set the check interval to be 1/5 of the estimated time to
	 * send a single character, and make it at least 1.  The check
	 * interval should also be less than the timeout.
	 *
	 * Note: we have to use pretty tight timings here to satisfy
	 * the NIST-PCTS.
	 */
	char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
	char_time = char_time / 5;

	if (char_time == 0)
		char_time = 1;

	if (timeout && timeout < char_time)
		char_time = timeout;
	/*
	 * If the transmitter hasn't cleared in twice the approximate
	 * amount of time to send the entire FIFO, it probably won't
	 * ever clear.  This assumes the UART isn't doing flow
	 * control, which is currently the case.  Hence, if it ever
	 * takes longer than info->timeout, this is probably due to a
	 * UART bug of some kind.  So, we clamp the timeout parameter at
	 * 2*info->timeout.
	 */

	if (!timeout || timeout > 2*info->timeout)
		timeout = 2*info->timeout;

	while (!((lsr = PTLSER_READ_REG(info->base+ UART_LSR)) & UART_LSR_TEMT)) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(char_time);

		if (signal_pending(current))
			break;

		if (timeout && time_after(jiffies, orig_jiffies + timeout))
			break;
	}
	set_current_state(TASK_RUNNING);
}
#endif

/*
 * This routine is called by tty_hangup() when a hangup is signaled.
 */
void ptlser_hangup(struct tty_struct * tty)
{
	struct ptlser_struct * info = (struct ptlser_struct *)tty->driver_data;
	ptlser_flush_buffer(tty);
	ptlser_shutdown(info);
	info->event = 0;
	info->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}


/*
 * ptlser_break_ctl() --- routine which turns the break handling on or off
 */
static int ptlser_break_ctl(struct tty_struct *tty, int break_state)
{
	struct ptlser_struct * info = (struct ptlser_struct *)tty->driver_data;
	PTLSER_LOCK_INIT();
	PTLSER_LOCK(&info->slock);
	if (break_state == -1){
		PTLSER_WRITE_REG(PTLSER_READ_REG(info->base + UART_LCR) | UART_LCR_SBC, info->base + UART_LCR);
	} else {
		PTLSER_WRITE_REG(PTLSER_READ_REG(info->base + UART_LCR) & ~UART_LCR_SBC, info->base + UART_LCR);
	}
	PTLSER_UNLOCK(&info->slock);
	return 0;
}

/*
 * This is the serial driver's generic interrupt routine
 */

static irqreturn_t ptlser_interrupt(int irq, void *dev_id)
{
	int			status, iir, i;
	struct ptlser_struct *	info;
	struct ptlser_struct *	port;
	int			max, msr;
	int			pass_counter = 0;
	int			int_cnt;
	int			handled=0;
	int			vect_flag;

	port = 0;

        for (i=0; i < PTLSER_BOARDS; i++){
            if (dev_id == &(ptlser_var_table[i*PTLSER_PORTS_PER_BOARD])){
                port = dev_id;
                break;
            }
        }

        if (i == PTLSER_BOARDS)
            goto irq_stop;

        if (port == 0)
            goto irq_stop;

        max = ptlser_numports[ptlsercfg[i].board_type-1];
	pass_counter = 0;

	do {
		vect_flag = 0;

		for (i = 0; i < max; i++){
			info = port + i;
			int_cnt = 0;

			do{
				iir = PTLSER_READ_REG(info->base+UART_IIR);
				if (iir == PTLSER_ISR_NO_INT_PENDING){
					vect_flag++;
					break;
				}

				handled = 1;

				spin_lock(&info->slock);
				status = PTLSER_READ_REG(info->base+UART_LSR);

				if (iir & PTLSER_ISR_RX_DATA){
					status &= info->read_status_mask;
					if ( status & UART_LSR_DR )
						ptlser_receive_chars(info, &status);
				}

				msr = PTLSER_READ_REG(info->base + UART_MSR);
				if (msr & UART_MSR_ANY_DELTA ) {
					ptlser_check_modem_status(info, msr);
				}

				if (iir & PTLSER_ISR_TX_EMPTY){
					if ( status & UART_LSR_THRE ) {
						ptlser_transmit_chars(info);
					}
				}

				spin_unlock(&info->slock);
			} while ( int_cnt++ < PTLSER_ISR_PASS_LIMIT);
		}

		if(vect_flag == max)
			break;

	} while ( pass_counter++ < PTLSER_ISR_PASS_LIMIT);
	
irq_stop:
	return IRQ_RETVAL(handled);	
}


#if defined(_SCREEN_INFO_H)
static void ptlser_receive_chars(struct ptlser_struct *info, int *status)
{
	struct tty_struct *	tty = info->tty;
	unsigned char	ch, gdl=0;
	int			    cnt = 0;
	int 			recv_room;
	int			    max = 256;
	unsigned long 	flags;

	if ( *status & UART_LSR_SPECIAL )
		goto intr_old;
	 
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	recv_room = tty_buffer_request_room(tty, PTLSER_FIFO_SIZE);
#else
	recv_room = tty_buffer_request_room((struct tty_port *) info, PTLSER_FIFO_SIZE);
#endif

	if(recv_room){
		gdl = getRxCount(info);
		if(gdl > recv_room)
			gdl = recv_room;

		if(gdl){
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
			tty_insert_flip_string(tty, info->base + PTLSER_FIFO_D_OFFSET, gdl);
#else
			tty_insert_flip_string((struct tty_port *) info, info->base + PTLSER_FIFO_D_OFFSET, gdl);
#endif
			cnt = gdl;
		}
	}
	else{
		set_bit(TTY_THROTTLED, &tty->flags);
	}

	goto end_intr;

intr_old:

	do {
		if(max-- <0)
			break;

		ch = PTLSER_READ_REG(info->base + UART_RX);

		if ( *status & UART_LSR_SPECIAL ) {
			if ( *status & UART_LSR_BI ) {
				flags = TTY_BREAK;
				info->icount.brk++;

				if ( info->flags & ASYNC_SAK )
					do_SAK(tty);
			} else if ( *status & UART_LSR_PE ) {
				flags = TTY_PARITY;
				info->icount.parity++;
			} else if ( *status & UART_LSR_FE ) {
				flags = TTY_FRAME;			
				info->icount.frame++;
			} else if ( *status & UART_LSR_OE ) {
				flags = TTY_OVERRUN;
				info->icount.overrun++;
			} else
				flags = TTY_BREAK;

		} else
			flags = 0;
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
		tty_insert_flip_char(tty, ch, flags);
#else
		tty_insert_flip_char((struct tty_port *) info, ch, flags);
#endif
		cnt++;

		*status = PTLSER_READ_REG(info->base+UART_LSR);
	} while ( *status & UART_LSR_DR );

end_intr:	// add by Victor Yu. 09-02-2002

#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	tty_flip_buffer_push(tty);
#else
	tty_flip_buffer_push((struct tty_port *) info);
#endif
}
#else
static void ptlser_receive_chars(struct ptlser_struct *info,
					 int *status)
{
	struct tty_struct *	tty = info->tty;
	unsigned char		ch, gdl;
	int			cnt = 0;
	unsigned char           *cp;
	char                    *fp;
	int 			recv_room;
	int			max = 256;
	
        cp = tty->flip.char_buf;
        fp = tty->flip.flag_buf;
	if ( *status & UART_LSR_SPECIAL )
		goto intr_old;

	recv_room = tty->ldisc.receive_room(tty);

	if(recv_room){
		gdl = getRxCount(info);
		if(gdl > recv_room)
			gdl = recv_room;
		if(gdl){
			cnt = gdl;
			memcpy(cp,info->base + PTLSER_FIFO_D_OFFSET, cnt);
			memset(fp, TTY_NORMAL, cnt);
			cp+=cnt;
			fp+=cnt;
		}
	}
	else{
		set_bit(TTY_THROTTLED, &tty->flags);
	}

	goto end_intr;

intr_old:
	do {
		if(max-- <0)
			break;

		ch = PTLSER_READ_REG(info->base + UART_RX);

		if ( *status & UART_LSR_SPECIAL ) {
			if ( *status & UART_LSR_BI ) {
				*fp++ = TTY_BREAK;
				info->icount.brk++;
				if ( info->flags & ASYNC_SAK )
					do_SAK(tty);
			} else if ( *status & UART_LSR_PE ) {
				*fp++ = TTY_PARITY;
				info->icount.parity++;
			} else if ( *status & UART_LSR_FE ) {
				*fp++ = TTY_FRAME;
				info->icount.frame++;
			} else if ( *status & UART_LSR_OE ) {
				*fp++ = TTY_OVERRUN;
				info->icount.overrun++;
			} else
				*fp++ = 0;
		} else
			*fp++ = 0;

		*cp++ = ch;
		cnt++;

		*status = PTLSER_READ_REG(info->base+UART_LSR);
	} while ( *status & UART_LSR_DR );

end_intr:

	tty->ldisc.receive_buf(tty, tty->flip.char_buf, tty->flip.flag_buf, cnt);

}
#endif


static void ptlser_transmit_chars(struct ptlser_struct *info)
{
	int	count, cnt;
	int tx_cnt;

	if ( info->xmit_buf == 0 ){
	    return;
	}

	if (info->xmit_cnt == 0) {
		if ( info->xmit_cnt < WAKEUP_CHARS ) {
			set_bit(PTLSER_EVENT_TXLOW,&info->event);
    			schedule_work(&info->tqueue);
		}
	}


	if (info->tty->stopped){
		info->IER &= ~UART_IER_THRI;
		PTLSER_WRITE_REG(info->IER, info->base + UART_IER);
		return;
	}

	cnt = info->xmit_cnt;
	tx_cnt = PTLSER_FIFO_SIZE - getTxCount(info);
	cnt = MIN(info->xmit_cnt, tx_cnt);
	if(cnt){
		memcpy(info->base + PTLSER_FIFO_D_OFFSET, info->xmit_buf+info->xmit_tail, cnt);
		info->xmit_tail += cnt;
		info->xmit_tail &= (SERIAL_XMIT_SIZE - 1);
		info->xmit_cnt -= cnt;
	}

    	info->icount.tx += cnt;

	if ( info->xmit_cnt < WAKEUP_CHARS ) {
		set_bit(PTLSER_EVENT_TXLOW,&info->event);
    		schedule_work(&info->tqueue);
	}


}

static void ptlser_check_modem_status(struct ptlser_struct *info,
					      int status)
{
	/* update input line counters */
	if ( status & UART_MSR_TERI )
	    info->icount.rng++;
	if ( status & UART_MSR_DDSR )
	    info->icount.dsr++;
	if ( status & UART_MSR_DDCD )
	    info->icount.dcd++;
	if ( status & UART_MSR_DCTS )
	    info->icount.cts++;

	wake_up_interruptible(&info->delta_msr_wait);

	if ( (info->flags & ASYNC_CHECK_CD) && (status & UART_MSR_DDCD) ) {
		if ( status & UART_MSR_DCD )
			wake_up_interruptible(&info->open_wait);
		else if ( !((info->flags & ASYNC_CALLOUT_ACTIVE) && (info->flags & ASYNC_CALLOUT_NOHUP)) )
	        set_bit(PTLSER_EVENT_HANGUP,&info->event);
    		schedule_work(&info->tqueue);
	}

	if ( info->flags & ASYNC_CTS_FLOW ) {
		if ( info->tty->hw_stopped ) {
			if (status & UART_MSR_CTS ){
			    	info->tty->hw_stopped = 0;
		       		set_bit(PTLSER_EVENT_TXLOW,&info->event);
    	            		schedule_work(&info->tqueue);
	        	}
		} else {
			if ( !(status & UART_MSR_CTS) ){
			    	info->tty->hw_stopped = 1;
			}
		}
	}
}

static int ptlser_block_til_ready(struct tty_struct *tty, struct file * filp, struct ptlser_struct *info)
{
	DECLARE_WAITQUEUE(wait, current);
	int			retval;
	int			do_clocal = 0;
	PTLSER_LOCK_INIT();

	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ( (filp->f_flags & O_NONBLOCK) || (tty->flags & (1 << TTY_IO_ERROR)) ) {
		if ( info->flags & ASYNC_CALLOUT_ACTIVE )
			return -EBUSY;

		info->flags |= ASYNC_NORMAL_ACTIVE;

		return 0;
	}

	if ( info->flags & ASYNC_CALLOUT_ACTIVE ) {
		if ( info->normal_termios.c_cflag & CLOCAL )
			do_clocal = 1;
	} else {
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))	
		if ( tty->termios->c_cflag & CLOCAL )
#else
		if ( tty->termios.c_cflag & CLOCAL )
#endif
			do_clocal = 1;
	}

	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, info->count is dropped by one, so that
	 * ptlser_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
	
	PTLSER_LOCK(&info->slock);
	if ( !tty_hung_up_p(filp) )
	    info->count--;
	PTLSER_UNLOCK(&info->slock);

	info->blocked_open++;

	while ( 1 ) {

		PTLSER_LOCK(&info->slock);
		if ( !(info->flags & ASYNC_CALLOUT_ACTIVE) )
			PTLSER_WRITE_REG(PTLSER_READ_REG(info->base + UART_MCR) | UART_MCR_DTR | UART_MCR_RTS,
		     info->base + UART_MCR);
		PTLSER_UNLOCK(&info->slock);

		set_current_state(TASK_INTERRUPTIBLE);

		if ( tty_hung_up_p(filp) || !(info->flags & ASYNC_INITIALIZED) ) {
#ifdef SERIAL_DO_RESTART
			if ( info->flags & ASYNC_HUP_NOTIFY )
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;
#else
			retval = -EAGAIN;
#endif
			break;
		}

		if ( !(info->flags & ASYNC_CALLOUT_ACTIVE) && !(info->flags & ASYNC_CLOSING) &&
		 (do_clocal || (PTLSER_READ_REG(info->base + UART_MSR) & UART_MSR_DCD)) )

			break;
		if ( signal_pending(current) ) {
			retval = -ERESTARTSYS;
			break;
		}

		schedule();
	}

	set_current_state(TASK_RUNNING);
	remove_wait_queue(&info->open_wait, &wait);

	if ( !tty_hung_up_p(filp) )
		info->count++;

	info->blocked_open--;

	if ( retval )
		return retval;

	info->flags |= ASYNC_NORMAL_ACTIVE;

	return 0;
}

static int ptlser_startup(struct ptlser_struct * info)
{
	unsigned long page;
	unsigned char reg_flag;
	PTLSER_LOCK_INIT();

	page = GET_FPAGE(GFP_KERNEL, 0); //simoncc
	if ( !page )
	    return -ENOMEM;

	PTLSER_LOCK(&info->slock);

	if ( info->flags & ASYNC_INITIALIZED ) {
		free_pages(page, 0);
		PTLSER_UNLOCK(&info->slock);

		return 0;
	}

	if ( !info->base || !info->type ) {
		if ( info->tty )
			set_bit(TTY_IO_ERROR, &info->tty->flags);
		free_pages(page, 0);
		PTLSER_UNLOCK(&info->slock);

		return 0;
	}

	if ( info->xmit_buf )
		free_pages(page, 0);
	else
		info->xmit_buf = (unsigned char *)page;

	ptlser_change_speed(info, 0);

	/*
	 * Clear the FIFO buffers and disable them
	 * (they will be reenabled in ptlser_change_speed())
	 */
	//simoncc ++++++++++++++++++++
//	PTLSER_WRITE_REG((PTLSER_FCR_RX_FLUSH | PTLSER_FCR_TX_FLUSH),info->base + UART_FCR);
//	PTLSER_WRITE_REG((PTLSER_FCR_RX_FLUSH | PTLSER_FCR_TX_FLUSH),info->base + UART_FCR);
	//simoncc --------------------
	/*
	 * At this point there's no way the LSR could still be 0xFF;
	 * if it is, then bail out, because there's likely no UART
	 * here.
	 */

	if ( PTLSER_READ_REG(info->base + UART_LSR) == 0xff ) {
		PTLSER_UNLOCK(&info->slock);

		if (capable(CAP_SYS_ADMIN)) {
			if ( info->tty )
				set_bit(TTY_IO_ERROR, &info->tty->flags);

			return(0);
		} else
			return -ENODEV;
	}

	/*
	 * Clear the interrupt registers.
	 */
	(void)PTLSER_READ_REG(info->base + UART_LSR);
	(void)PTLSER_READ_REG(info->base + UART_RX);
	(void)PTLSER_READ_REG(info->base + UART_IIR);
	(void)PTLSER_READ_REG(info->base + UART_MSR);

	/*
	 * Now, initialize the UART
	 */
	PTLSER_WRITE_REG(UART_LCR_WLEN8, info->base + UART_LCR);	/* reset DLAB */
	info->MCR = UART_MCR_DTR | UART_MCR_RTS;
	PTLSER_WRITE_REG(info->MCR, info->base + UART_MCR);

	/*
	 * Initialize enhance mode register
	 */
	reg_flag = PTLSER_EFR_ENHANCE_MODE;
	PTLSER_WRITE_REG(reg_flag, info->base + PTLSER_EFR_OFFSET);

	reg_flag = PTLSER_SFR_950 | PTLSER_SFR_TX_FIFO_COUNT;
	PTLSER_WRITE_REG(reg_flag, info->base + PTLSER_SFR_OFFSET);

	PTLSER_WRITE_REG(1, info->base + PTLSER_TTL_OFFSET);
	PTLSER_WRITE_REG(96, info->base + PTLSER_RTL_OFFSET); //96
	PTLSER_WRITE_REG(16, info->base + PTLSER_FCL_OFFSET);
	PTLSER_WRITE_REG(110, info->base + PTLSER_FCH_OFFSET);

	/*
	 * Finally, enable interrupts
	 */
	info->IER = UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI;
	PTLSER_WRITE_REG(info->IER, info->base + UART_IER); /* enable interrupts */

	/*
	 * And clear the interrupt registers again for luck.
	 */
	(void)PTLSER_READ_REG(info->base + UART_LSR);
	(void)PTLSER_READ_REG(info->base + UART_RX);
	(void)PTLSER_READ_REG(info->base + UART_IIR);
	(void)PTLSER_READ_REG(info->base + UART_MSR);

	if ( info->tty )
		test_and_clear_bit(TTY_IO_ERROR, &info->tty->flags);

	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;

	/*
	 * and set the speed of the serial port
	 */
	ptlser_change_speed(info, 0);
	PTLSER_UNLOCK(&info->slock);

	info->flags |= ASYNC_INITIALIZED;

	return 0;
}

/*
 * This routine will shutdown a serial port; interrupts maybe disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 */
static void ptlser_shutdown(struct ptlser_struct * info)
{
	unsigned char reg_flag;

	PTLSER_LOCK_INIT();

	if ( !(info->flags & ASYNC_INITIALIZED) )
		return;

	PTLSER_LOCK(&info->slock);

	/*
	 * clear delta_msr_wait queue to avoid mem leaks: we may free the irq
	 * here so the queue might never be waken up
	 */
	wake_up_interruptible(&info->delta_msr_wait);

	/*
	 * Free the IRQ, if necessary
	 */
	if ( info->xmit_buf ) {
		free_pages((unsigned long)info->xmit_buf,0); //simoncc
		info->xmit_buf = 0;
	}

	reg_flag = 0;
	PTLSER_WRITE_REG(reg_flag, info->base + PTLSER_EFR_OFFSET);
	PTLSER_WRITE_REG(reg_flag, info->base + PTLSER_SFR_OFFSET);

	info->IER = 0;
	PTLSER_WRITE_REG(0x00, info->base + UART_IER);
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	if ( !info->tty || (info->tty->termios->c_cflag & HUPCL) )
#else
	if ( !info->tty || (info->tty->termios.c_cflag & HUPCL) )
#endif
		info->MCR &= ~(UART_MCR_DTR | UART_MCR_RTS);
	PTLSER_WRITE_REG(info->MCR, info->base + UART_MCR);

	/* read data port to reset things */
	(void)PTLSER_READ_REG(info->base + UART_RX);

	if ( info->tty )
		set_bit(TTY_IO_ERROR, &info->tty->flags);

	info->flags &= ~ASYNC_INITIALIZED;

	PTLSER_UNLOCK(&info->slock);
}

/*
 * This routine is called to set the UART divisor registers to match
 * the specified baud rate for a serial port.
 */
static int ptlser_change_speed(struct ptlser_struct *info, struct ktermios *old_termios)
{
	unsigned	cflag, cval, fcr;
        int             ret = 0;
	long baud;
	int reg_flag;

#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	if ( !info->tty || !info->tty->termios )
		return ret;

	cflag = info->tty->termios->c_cflag;
#else	
	if ( !info->tty )
		return ret;

	cflag = info->tty->termios.c_cflag;
#endif

	if ( !(info->base) )
		return ret;

#ifndef B921600
#define B921600 (B460800 +1)
#endif
	switch( cflag & (CBAUD | CBAUDEX) ){
        case B921600 : info->speed = baud = 921600; break;
        case B460800 : info->speed = baud = 460800; break;
        case B230400 : info->speed = baud = 230400; break;
        case B115200 : info->speed = baud = 115200; break;
        case B57600 : info->speed = baud = 57600; break;
        case B38400 : info->speed = baud = 38400;break;
        case B19200 : info->speed = baud = 19200; break;
        case B9600 : info->speed = baud = 9600; break;
        case B4800 : info->speed = baud = 4800; break;
        case B2400 : info->speed = baud = 2400; break;
        case B1800 : info->speed = baud = 1800; break;
        case B1200 : info->speed = baud = 1200; break;
        case B600 : info->speed = baud = 600; break;
        case B300 : info->speed = baud = 300; break;
        case B200 : info->speed = baud = 200; break;
        case B150 : info->speed = baud = 150; break;
        case B134 : info->speed = baud = 134; break;
        case B110 : info->speed = baud = 110; break;
        case B75 : info->speed = baud = 75; break;
        case B50 : info->speed = baud = 50; break;
        default: info->speed = baud = 0; break;
	}

    ptlser_set_baud(info, baud);

	/* byte size and parity */
	switch ( cflag & CSIZE ) {
		case CS5: cval = 0x00; break;
		case CS6: cval = 0x01; break;
		case CS7: cval = 0x02; break;
		case CS8: cval = 0x03; break;
		default:  cval = 0x00; break;	/* too keep GCC shut... */
	}

	if ( cflag & CSTOPB )
		cval |= 0x04;

	if ( cflag & PARENB )
		cval |= UART_LCR_PARITY;

#ifndef CMSPAR
#define	CMSPAR 010000000000
#endif
	if ( !(cflag & PARODD) )
		cval |= UART_LCR_EPAR;

	if ( cflag & CMSPAR )
		cval |= UART_LCR_SPAR;

	fcr = UART_FCR_ENABLE_FIFO | PTLSER_FCR_DMA_ENABLE;

	/* CTS flow control flag and modem status interrupts */
	info->IER &= ~UART_IER_MSI;
//	info->MCR &= ~UART_MCR_AFE;

	reg_flag = PTLSER_READ_REG(info->base + PTLSER_EFR_OFFSET);

	if ( cflag & CRTSCTS ) {
		info->flags |= ASYNC_CTS_FLOW;
		info->IER |= UART_IER_MSI;
		reg_flag |= (PTLSER_EFR_AUTO_RTS | PTLSER_EFR_AUTO_CTS);
	} else {
		info->flags &= ~ASYNC_CTS_FLOW;
		reg_flag &= ~(PTLSER_EFR_AUTO_RTS | PTLSER_EFR_AUTO_CTS);
	}

	PTLSER_WRITE_REG(info->MCR, info->base + UART_MCR);

	if ( cflag & CLOCAL ){
		info->flags &= ~ASYNC_CHECK_CD;
	}else {
		info->flags |= ASYNC_CHECK_CD;
		info->IER |= UART_IER_MSI;
	}

	PTLSER_WRITE_REG(info->IER, info->base + UART_IER);

	/*
	 * Set up parity check flag
	 */
	info->read_status_mask = UART_LSR_OE | UART_LSR_THRE | UART_LSR_DR;

	if ( I_INPCK(info->tty) )
		info->read_status_mask |= UART_LSR_FE | UART_LSR_PE;

	if ( I_BRKINT(info->tty) || I_PARMRK(info->tty) )
		info->read_status_mask |= UART_LSR_BI;

	info->ignore_status_mask = 0;

	if ( I_IGNBRK(info->tty) ) {
		info->ignore_status_mask |= UART_LSR_BI;
		info->read_status_mask |= UART_LSR_BI;
	    /*
	     * If we're ignore parity and break indicators, ignore
	     * overruns too.  (For real raw support).
	     */
		if ( I_IGNPAR(info->tty) ) {
			info->ignore_status_mask |= UART_LSR_OE|UART_LSR_PE|UART_LSR_FE;
			info->read_status_mask |= UART_LSR_OE|UART_LSR_PE|UART_LSR_FE;
		}
	}

	PTLSER_WRITE_REG(START_CHAR(info->tty), info->base + PTLSER_XON1_OFFSET);
	PTLSER_WRITE_REG(START_CHAR(info->tty), info->base + PTLSER_XON2_OFFSET);
	PTLSER_WRITE_REG(STOP_CHAR(info->tty), info->base + PTLSER_XOFF1_OFFSET);
	PTLSER_WRITE_REG(STOP_CHAR(info->tty), info->base + PTLSER_XOFF2_OFFSET);

	if ( I_IXON(info->tty) )
		reg_flag |= PTLSER_EFR_INBAND_RX_MODE_1;
	else
		reg_flag &= ~PTLSER_EFR_INBAND_RX_MODE_1;

	if ( I_IXOFF(info->tty) )
		reg_flag |= PTLSER_EFR_INBAND_TX_MODE_1;
	else
		reg_flag &= ~PTLSER_EFR_INBAND_TX_MODE_1;

	PTLSER_WRITE_REG(reg_flag, info->base + PTLSER_EFR_OFFSET);
	PTLSER_WRITE_REG(fcr, info->base + UART_FCR);		    /* set fcr */
	PTLSER_WRITE_REG(cval, info->base + UART_LCR);  
	
        return ret;
}


static int ptlser_set_baud(struct ptlser_struct *info, long newspd)
{
	int		i;
	int		quot = 0;
	unsigned char	cval;
        int             ret = 0;

#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	if ( !info->tty || !info->tty->termios )
		return ret;
#else
	if ( !info->tty )
		return ret;
#endif

	if ( !(info->base) )
		return ret;

	if ( newspd > info->MaxCanSetBaudRate )
		return 0;

	info->realbaud = newspd;

	for ( i=0; i<BAUD_TABLE_NO && newspd != ptlser_baud_table[i]; i++ );

	if ( i == BAUD_TABLE_NO ){
		quot = info->baud_base / info->speed;
		if ( info->speed <= 0 || info->speed > info->MaxCanSetBaudRate )
			quot = 0;
	}else{
		if ( newspd == 134 ) {
			quot = (2 * info->baud_base / 269);
			info->speed = 134;
		} else if ( newspd ) {
			quot = info->baud_base / newspd;

			if(quot==0)
				quot = 1;
		} else {
			quot = 0;
		}
	}

	info->timeout = (int)((unsigned int)(info->xmit_fifo_size*HZ*10*quot) / (unsigned int)info->baud_base);
	info->timeout += HZ/50;		/* Add .02 seconds of slop */

	if ( quot ) {
		info->MCR |= UART_MCR_DTR;
		PTLSER_WRITE_REG(info->MCR, info->base + UART_MCR);
	} else {
		info->MCR &= ~UART_MCR_DTR;
		PTLSER_WRITE_REG(info->MCR, info->base + UART_MCR);

		return ret;
	}

	cval = PTLSER_READ_REG(info->base + UART_LCR);
	PTLSER_WRITE_REG(cval | UART_LCR_DLAB, info->base + UART_LCR);  /* set DLAB */
	PTLSER_WRITE_REG(quot & 0xff, info->base + UART_DLL);	    /* LS of divisor */
	PTLSER_WRITE_REG(quot >> 8, info->base + UART_DLM); 	    /* MS of divisor */
	PTLSER_WRITE_REG(cval, info->base + UART_LCR);		    /* reset DLAB */
	// simoncc patch...
	if (info->realbaud < 9600) {				
		PTLSER_WRITE_REG(3, info->base + PTLSER_RTL_OFFSET);			
	} else if (info->realbaud < 38400) {
		PTLSER_WRITE_REG(8, info->base + PTLSER_RTL_OFFSET);
	} else if (info->realbaud < 115200) {
		PTLSER_WRITE_REG(14, info->base + PTLSER_RTL_OFFSET);
	} else {
		PTLSER_WRITE_REG(96, info->base + PTLSER_RTL_OFFSET);
	}

	if ( i == BAUD_TABLE_NO ){
		quot = info->baud_base % info->speed;
		quot *= 8;
		if ( (quot % info->speed) > (info->speed / 2) ) {
			quot /= info->speed;
			quot++;
		} else {
			quot /= info->speed;
		}
	}

    return ret;
}



/*
 * ------------------------------------------------------------
 * friends of ptlser_ioctl()
 * ------------------------------------------------------------
 */
static int ptlser_get_serial_info(struct ptlser_struct * info, struct serial_struct * retinfo)
{
	struct serial_struct	tmp;

	if ( !retinfo )
		return(-EFAULT);

	memset(&tmp, 0, sizeof(tmp));
	tmp.type = info->type;
	tmp.line = info->port;
	tmp.port = *info->base;
	tmp.irq = info->irq;
	tmp.flags = info->flags;
	tmp.baud_base = info->baud_base;
	tmp.close_delay = info->close_delay;
	tmp.closing_wait = info->closing_wait;
	tmp.custom_divisor = info->custom_divisor;
	tmp.hub6 = 0;

	if(copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
		return -EFAULT;

	return 0;
}

static int ptlser_set_serial_info(struct ptlser_struct * info, struct serial_struct * new_info)
{
	struct serial_struct	new_serial;
	unsigned int		flags;
	int			retval = 0;

	PTLSER_LOCK_INIT();

	if ( !new_info || !info->base )
		return -EFAULT;

	if(copy_from_user(&new_serial, new_info, sizeof(new_serial)))
		return -EFAULT;

	if ( (new_serial.irq != info->irq) ||
	     (new_serial.port != *info->base) )
		return -EPERM;

	flags = info->flags & ASYNC_SPD_MASK;

       if ( !capable(CAP_SYS_ADMIN)) {
		if ( (new_serial.baud_base != info->baud_base) ||
		     (new_serial.close_delay != info->close_delay) ||
		    ((new_serial.flags & ~ASYNC_USR_MASK) !=
		    (info->flags & ~ASYNC_USR_MASK)) )
			return(-EPERM);

		info->flags = ((info->flags & ~ASYNC_USR_MASK) | (new_serial.flags & ASYNC_USR_MASK));
	} else {
	    /*
	     * OK, past this point, all the error checking has been done.
	     * At this point, we start making changes.....
	     */
		info->flags = ((info->flags & ~ASYNC_FLAGS) | (new_serial.flags & ASYNC_FLAGS));
		info->close_delay = new_serial.close_delay * HZ/100;
		info->closing_wait = new_serial.closing_wait * HZ/100;

		if( (new_serial.baud_base != info->baud_base) ||
		    (new_serial.custom_divisor != info->custom_divisor) )
			info->custom_baud_rate = new_serial.baud_base/new_serial.custom_divisor;
	}

	info->type = new_serial.type;
	info->xmit_fifo_size = PTLSER_FIFO_SIZE;

	if ( info->flags & ASYNC_INITIALIZED ) {
		if ( flags != (info->flags & ASYNC_SPD_MASK) ){
			PTLSER_LOCK(&info->slock);
			ptlser_change_speed(info,0);
			PTLSER_UNLOCK(&info->slock);
		}
	} else{
	    retval = ptlser_startup(info);
	}

	return retval;
}

/*
 * ptlser_get_lsr_info - get line status register info
 *
 * Purpose: Let user call ioctl() to get info when the UART physically
 *	    is emptied.  On bus types like RS485, the transmitter must
 *	    release the bus after transmitting. This must be done when
 *	    the transmit shift register is empty, not be done when the
 *	    transmit holding register is empty.  This functionality
 *	    allows an RS485 driver to be written in user space.
 */
static int ptlser_get_lsr_info(struct ptlser_struct * info, unsigned int *value)
{
	unsigned char	status;
	unsigned int	result;
	PTLSER_LOCK_INIT();

	PTLSER_LOCK(&info->slock);
	status = PTLSER_READ_REG(info->base + UART_LSR);
	PTLSER_UNLOCK(&info->slock);
	result = ((status & UART_LSR_TEMT) ? TIOCSER_TEMT : 0);
	put_to_user(result, value);

	return 0;
}

/*
 * This routine sends a break character out the serial port.
 */
static void ptlser_send_break(struct ptlser_struct * info, int duration)
{
	PTLSER_LOCK_INIT();

	if ( !info->base )
		return;

	set_current_state(TASK_INTERRUPTIBLE);

	PTLSER_LOCK(&info->slock);
	PTLSER_WRITE_REG(PTLSER_READ_REG(info->base + UART_LCR) | UART_LCR_SBC, info->base + UART_LCR);
	PTLSER_UNLOCK(&info->slock);

	schedule_timeout(duration);

	PTLSER_LOCK(&info->slock);
	PTLSER_WRITE_REG(PTLSER_READ_REG(info->base + UART_LCR) & ~UART_LCR_SBC, info->base + UART_LCR);
	PTLSER_UNLOCK(&info->slock);
	
	set_current_state(TASK_RUNNING);
}


static int ptlser_tiocmget(struct tty_struct *tty, struct file *file)
{
	struct ptlser_struct *info = (struct ptlser_struct *) tty->driver_data;
	unsigned char control, status;
	PTLSER_LOCK_INIT();

	if (PORTNO(tty) == PTLSER_PORTS)
		return -ENOIOCTLCMD;

	if (tty->flags & (1 << TTY_IO_ERROR))
		return -EIO;

	control = info->MCR;
	
	PTLSER_LOCK(&info->slock);
	status = PTLSER_READ_REG(info->base + UART_MSR);

	if (status & UART_MSR_ANY_DELTA)
		ptlser_check_modem_status(info, status);

	PTLSER_UNLOCK(&info->slock);

	return ((control & UART_MCR_RTS) ? TIOCM_RTS : 0) |
	    ((control & UART_MCR_DTR) ? TIOCM_DTR : 0) |
	    ((status & UART_MSR_DCD) ? TIOCM_CAR : 0) |
	    ((status & UART_MSR_RI) ? TIOCM_RNG : 0) |
	    ((status & UART_MSR_DSR) ? TIOCM_DSR : 0) |
	    ((status & UART_MSR_CTS) ? TIOCM_CTS : 0);
}

static int ptlser_tiocmset(struct tty_struct *tty, struct file *file, unsigned int set, unsigned int clear)
{
	struct ptlser_struct *info = (struct ptlser_struct *) tty->driver_data;
	PTLSER_LOCK_INIT();

	if (PORTNO(tty) == PTLSER_PORTS)
		return (-ENOIOCTLCMD);

	if (tty->flags & (1 << TTY_IO_ERROR))
		return (-EIO);

	PTLSER_LOCK(&info->slock);

	if (set & TIOCM_RTS)
		info->MCR |= UART_MCR_RTS;
	if (set & TIOCM_DTR)
		info->MCR |= UART_MCR_DTR;
	if (clear & TIOCM_RTS)
		info->MCR &= ~UART_MCR_RTS;
	if (clear & TIOCM_DTR)
		info->MCR &= ~UART_MCR_DTR;

	PTLSER_WRITE_REG(info->MCR, info->base + UART_MCR);
	PTLSER_UNLOCK(&info->slock);

	return 0;
}




/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int ptlser_initHw(int board,struct ptlser_hwconf *hwconf)
{
	struct ptlser_struct *	info;
    int     retval;
	int	i,n;

	n = board*PTLSER_PORTS_PER_BOARD;
	info = &ptlser_var_table[n];
	printk(DRIVER_NAME "ttyPPU%d - ttyPPU%d\n", n, n+hwconf->ports-1);
	printk(DRIVER_NAME "max. baud rate = %d bps.\n", hwconf->MaxCanSetBaudRate[0]);
	
	for ( i=0; i<hwconf->ports; i++, n++, info++ ) {
		info->port = n;
		info->base = hwconf->ioaddr[i];
		info->irq = hwconf->irq;
		info->board_type = hwconf->board_type;
		info->flags = ASYNC_SHARE_IRQ;
		info->type = hwconf->uart_type;
		info->baud_base = hwconf->baud_base[i];
		info->MaxCanSetBaudRate = hwconf->MaxCanSetBaudRate[i];
		info->xmit_fifo_size = PTLSER_FIFO_SIZE;
		info->custom_divisor = hwconf->baud_base[i] * 16;
		info->close_delay = 5*HZ/10;
		info->closing_wait = 30*HZ;
		INIT_WORK(&info->tqueue, ptlser_do_softint);
		info->normal_termios = DRV_VAR_P(init_termios);
		init_waitqueue_head(&info->open_wait);
		init_waitqueue_head(&info->close_wait);
		init_waitqueue_head(&info->delta_msr_wait);
		info->speed = 9600;
		spin_lock_init(&info->slock);
		printk(DRIVER_NAME "ttyPPU%d c %d %d\n", n, ttymajor, n);
	}


	/*
	 * Allocate the IRQ if necessary
	 */
	

	/* before set INT ISR, disable all int */
	for(i=0; i<hwconf->ports; i++)
		PTLSER_WRITE_REG(PTLSER_READ_REG(hwconf->ioaddr[i] + UART_IER) & 0xf0, hwconf->ioaddr[i]+UART_IER);

	n = board*PTLSER_PORTS_PER_BOARD;
	info = &ptlser_var_table[n];
        retval = request_irq(hwconf->irq, ptlser_interrupt, IRQ_T(info), "ptlser", info);

	if ( retval ) {
	    printk(DRIVER_NAME "Board %d: %s", board, ptlser_brdname[hwconf->board_type-1]);
	    printk(DRIVER_NAME "Request irq fail,IRQ (%d) may be conflit with another device.\n",info->irq);
	    return retval;
	}
	
        return 0;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
 * boot-time initialization
 */
static int __init	ptlser_module_init(void)
{		
	int			    i, m, retval, b;
    int             ret1, ret2;
#ifdef CONFIG_PCI
	struct pci_dev	*pdev=NULL;
	int			    index;
	unsigned char	busnum,devnum;
#endif
	struct ptlser_hwconf	hwconf;

	printk(DRIVER_NAME "Pericom PI7C9X795x/PI7C9X895x UART Device Driver Version %s\n",PTLSER_VERSION);

	ptlser_var_sdriver = alloc_tty_driver(PTLSER_PORTS + 1);
	if (!ptlser_var_sdriver)
		return -ENOMEM;

	for(i=0; i<PTLSER_BOARDS; i++)
		ptlsercfg[i].board_type = -1;
	
	/* Initialize the tty_driver structure */
	DRV_VAR_P(magic)            = TTY_DRIVER_MAGIC;
	DRV_VAR_P(name)             = "ttyPPU";
	DRV_VAR_P(major)            = ttymajor;
	DRV_VAR_P(minor_start)      = 0;
	DRV_VAR_P(num)              = PTLSER_PORTS + 1;
	DRV_VAR_P(type)             = TTY_DRIVER_TYPE_SERIAL;
	DRV_VAR_P(subtype)          = SERIAL_TYPE_NORMAL;
	DRV_VAR_P(init_termios)     = tty_std_termios;
	DRV_VAR_P(init_termios.c_cflag) = B9600|CS8|CREAD|HUPCL|CLOCAL;
	DRV_VAR_P(flags)            = TTY_DRIVER_REAL_RAW;
	tty_set_operations(DRV_VAR, &ptlser_ops);
	DRV_VAR_P(ttys)             = ptlser_var_tty;
	DRV_VAR_P(termios)          = ptlser_var_termios;
#if (LINUX_VERSION_CODE < VERSION_CODE(3,3,0))
	DRV_VAR_P(termios_locked)   = ptlser_var_termios_locked;
#endif
	memset(ptlser_var_table, 0, PTLSER_PORTS * sizeof(struct ptlser_struct));
	memset(&hwconf, 0, sizeof(struct ptlser_hwconf));

	/* start finding PCI board here */
#ifdef CONFIG_PCI
	{		
		int n = (sizeof(ptlser_pcibrds) / sizeof(ptlser_pcibrds[0])) - 1;
		index = 0;
		b = 0;
		m = 0;

		while (b < n) {
		    pdev = pci_get_device(ptlser_pcibrds[b].vendor, ptlser_pcibrds[b].device, pdev);
			if(pdev==NULL){
				b++;
				continue;
			}

			hwconf.pciInfo.busNum = busnum = pdev->bus->number;
			hwconf.pciInfo.devNum = devnum = PCI_SLOT(pdev->devfn)<<3;
			hwconf.pciInfo.pdev = pdev;
			printk(DRIVER_NAME "Found PTLSER %s board(BusNo=%d,DevNo=%d)\n",ptlser_brdname[(int)(ptlser_pcibrds[b].driver_data)-1],busnum,devnum >> 3);
			index++;

			if ( m >= PTLSER_BOARDS) {
				printk(DRIVER_NAME "Too many boards find (maximum %d),board not configured\n",PTLSER_BOARDS);
			} else {
				if ( pci_enable_device(pdev) ) {
					printk(DRIVER_NAME "enable fail !\n");
					continue;
				}
				retval = ptlser_get_PCI_conf(busnum, devnum, (int)ptlser_pcibrds[b].driver_data, &hwconf);
				if (retval < 0) {
					if (retval == PTLSER_ERR_IRQ)
						printk(DRIVER_NAME "Invalid interrupt number,board not configured\n");
					else if (retval == PTLSER_ERR_IRQ_CONFLIT)
						printk(DRIVER_NAME "Invalid interrupt number,board not configured\n");
					else if (retval == PTLSER_ERR_VECTOR)
						printk(DRIVER_NAME "Invalid interrupt vector,board not configured\n");
					else if (retval == PTLSER_ERR_IOADDR)
						printk(DRIVER_NAME "Invalid I/O address,board not configured\n");
					continue;

				}
				
	    		ptlsercfg[m] = hwconf;
	    		//init ptlsercfg first, or ptlsercfg data is not correct on ISR.
	    		//ptlser_initHwwill hook ISR.
	    		if(ptlser_initHw(m,&hwconf)<0)
					continue;
				m++;
			}
		}
	}
#endif
	printk(DRIVER_NAME "Tty devices major number = %d, callout devices major number = %d\n",ttymajor,calloutmajor);

#if (LINUX_VERSION_CODE >= VERSION_CODE(3,7,0))        	
	for (m=0; m < PTLSER_BOARDS; m++) {
		for (i=0; i < PTLSER_PORTS_PER_BOARD; i++) {
			int port_idx = (m * PTLSER_PORTS_PER_BOARD) + i;
			tty_port_init(&ptlser_var_table[port_idx].ttyPort);
			tty_port_link_device(&ptlser_var_table[port_idx].ttyPort, ptlser_var_sdriver, port_idx);
		}
	}
#endif
        ret1 = 0;
        ret2 = 0;

        if ( !(ret1=tty_register_driver(DRV_VAR)) ){
	    return 0;	            
        } else
	    printk(DRIVER_NAME "Couldn't install PTLSER driver !\n");


        if (ret1 || ret2) {
		for (i=0; i<PTLSER_BOARDS; i++) {
			if(ptlsercfg[i].board_type == -1)
				continue;
			else {
				free_irq(ptlsercfg[i].irq, &ptlser_var_table[i*PTLSER_PORTS_PER_BOARD]);
		    //todo: release io, vector
	                }
		}

		return -1;
        }
	
	return 0;	
}




static void __exit	ptlser_module_exit(void)
{
	int i, err=0;

	struct ktermios *tp;
	void *p;

	if ((err |= tty_unregister_driver(DRV_VAR)))
		printk(DRIVER_NAME "Couldn't unregister serial driver\n");

	for (i = 0; i < DRV_VAR->num; i++) {
		tp = DRV_VAR->termios[i];
		if (tp) {
			DRV_VAR->termios[i] = NULL;
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))		
			/* the ktermios of tty is not get through alloc after kernel 3.7, so we do not
			  *  have to free this */
			kfree(tp);
#endif
		}
#if (LINUX_VERSION_CODE < VERSION_CODE(3,3,0))
		tp = DRV_VAR->termios_locked[i];
		if (tp) {
			DRV_VAR->termios_locked[i] = NULL;
			kfree(tp);
		}
#endif
		if (!(DRV_VAR->flags & TTY_DRIVER_DYNAMIC_DEV))
			tty_unregister_device(DRV_VAR, i);
	}
	p = DRV_VAR->ttys;
	DRV_VAR->ttys = NULL;
#if (LINUX_VERSION_CODE < VERSION_CODE(3,3,0))
	DRV_VAR->termios = DRV_VAR->termios_locked = NULL;
#else
	DRV_VAR->termios = NULL;
#endif
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))	
	kfree(p);
#endif
#if (LINUX_VERSION_CODE < VERSION_CODE(3,7,0))
	cdev_del(&DRV_VAR->cdev);
#endif

 
        for(i=0; i<PTLSER_BOARDS; i++){
		struct pci_dev *pdev;

		if(ptlsercfg[i].board_type == -1){
			continue;
		}
		else{
			pdev = ptlsercfg[i].pciInfo.pdev;
			free_irq(ptlsercfg[i].irq, &ptlser_var_table[i*PTLSER_PORTS_PER_BOARD]);

			if(pdev!=NULL){ //PCI
				iounmap(ptlsercfg[i].ioaddr[1]);
				release_mem_region(pci_resource_start(pdev, 1),	pci_resource_len(pdev, 1));
				release_region(pci_resource_start(pdev,2), pci_resource_len(pdev,2));
			}	
		}
    }
    printk(DRIVER_NAME "unregister successfully\n");
}


module_init(ptlser_module_init);
module_exit(ptlser_module_exit); 
